#ifndef __msd__
#define __msd__

#include <stdio.h>
#include "stock.h"

#define MAX_SEED_RECORD_SIZE	4096
#define MIN_SEED_RECORD_SIZE 512
#define MIN_BLOCKETTE_SIZE 5

/****  Seed Data Header blockette info ****/
#define  NET_LEN   2
#define  STA_LEN   5
#define  LOC_LEN   2
#define  CHAN_LEN  3 

#define MSD_FIXED_DATA_SIZE 	48

#define MSD_ACTIVITY_FLAGS		101
#define MSD_CHAN			102
#define MSD_CNT				103
#define MSD_DATAFORMAT			104
#define MSD_FD				105
#define MSD_FILENAME			106
#define MSD_FIRST_DATA_FRAME		107
#define MSD_FRAMES_PER_RECORD		108
#define MSD_HAS_B100			109
#define MSD_HAS_B1000			110
#define MSD_HAS_SEED_HEADER		111
#define MSD_IO_FLAGS			112
#define MSD_LEVEL			113
#define MSD_LOC				114
#define MSD_NET				115
#define MSD_NSAMP			116
#define MSD_OFFSET			117
#define MSD_QUALITY_FLAGS		118
#define MSD_RECORD_SIZE			119
#define MSD_SAMPRATE			120
#define MSD_SAMPRATE_FLAGS		121
#define MSD_SEQUENCE			122
#define MSD_SPARC_ORDER			123
#define MSD_STA				124
#define MSD_TIME			125
#define MSD_TIME_CORRECTION		126
#define MSD_MSEED			127
#define MSD_MSEED_NBYTES	        128
#define MSD_RECORD			129
#define MSD_RAWTIME			130
#define MSD_NOMINAL_SAMPRATE		131
#define MSD_USEC			132
#define MSD_TIMING_QUALITY		133
#define MSD_FRAME_COUNT			134
#define MSD_HAS_B1001			135
#define MSD_OPAQUE			136
#define MSD_HAS_OPAQUE			137
#define MSD_XNSAMP			138


#define MSD_BAD_CK_DNIB			0x01
#define MSD_BAD_NSAMP		        0x02
#define MSD_BAD_REVERSE_INTEGRATION	0x04
#define MSD_BAD_STEIM_CODE		0x08
#define MSD_UNSUPPORTED_STEIM_LEVEL	0x10
#define MSD_UNSUPPORTED_DATATYPE	0x20
#define MSD_NO_BLOCKETTE_1000		0x40

#define STEIM_BYTES_PER_FRAME 64
#define MAX_SAMPLES_PER_RECORD 6700    /* Seed reference pp 123 says 6601; 
					  this is for 4096 byte record */
typedef struct Stcdef {
    int             bits,
		    cnt,
		    code,
		    dnib,
		    mn,
		    mx,
		    mask,
		    sign,
		    extend;
}               Stcdef;

/* Information for Seed Data Header */
typedef struct SDH1 {
    int             seq;
    int		    quality;
    char            net[16];
    char            sta[16];
    char            loc[16];
    char            chan[16];
    int             year;
    unsigned int    doy,
		    hour,
		    minutes,
		    sec,
		    msec_x_10;
    unsigned int    nsamp;
    int             samprate_factor,
		    samprate_multiplier;
    unsigned int    activity_flags;
    unsigned int    io_flags;
    unsigned int    quality_flags;
    unsigned int    blks_following;
    int             time_correction;
    unsigned int    data_offset;   /* Steim frames start at this offset
				    * in record */
    unsigned int    next_blockette_offset;
    double          epoch,
		    samprate;      /* not actually in the blockette, but
				    * converted to a useful form */
    double	    nominal_samprate ; /* to put into the standard header */
}               SDH1;


typedef struct B100 {
    unsigned int    type;
    unsigned int    next_blockette_offset;
    unsigned int    flags;
}               B100;

typedef struct B1000 {
    unsigned int    type;
    unsigned int    next_blockette_offset;
    unsigned int    dataformat;
    unsigned int    sparc_order;
    unsigned int    log2_record_length;
}               B1000;

typedef struct B1001 {
    unsigned int    type;
    unsigned int    next_blockette_offset;
    unsigned int    timing_quality;
    int    	    usec ; 
    unsigned int    frame_count ; 
}               B1001;

typedef struct Msd_opaque { 
    unsigned int  record ; 
    unsigned int  sparc_order ; 
    unsigned char flags ; 
    Tbl *fields ; 
    unsigned char *data ; 
    unsigned int nbytes ; 
} Msd_opaque ; 

typedef struct Stcflags {
    unsigned int    sparc_order:1; /* file has sparc order? */
    unsigned int    has_seed_header:1;	/* record has standard seed
					     * header? */
    unsigned int    has_b100:1;    /* record has blockette 100 (samprate) */
    unsigned int    has_b1000:1;   /* record has blockette 1000 (format &
				    * record size) */
    unsigned int    has_b1001:1;   /* record has blockette 1001 (timing) */
    unsigned int    has_opaque:1;   /* record has blockette 2000 (opaque) */
    unsigned int    another:1;     /* internal flag for setting next
				    * blockette field in output
				    * blockettes */
    unsigned int    bad_sdh:1 ; 
    unsigned int    first:1 ; 	   /* flag for first data record (for msdnxt) */
    unsigned int    new_last_x:1 ;  /* for cstc */
    unsigned int    new_record:1 ;  /* for cstc */

    unsigned int    return_bad_records : 1 ; /* for msdd to see bad records */
    unsigned int    announce_other_blockettes : 1 ; /* for msdd */
}               Stcflags;

typedef struct Msd {
    /* optional miniseed blockettes */
    SDH1            sdh;
    B1000           b1000;
    B100            b100;
    B1001	    b1001;
    Tbl		   *opaque ;

    /* miniseed input/output files */
    char           *path;	       /* input/output filename */
    int             fd;	               /* input/output file descriptor */
    Hook	   *hook ; 	       /* private pointer */
    unsigned char  *mseed ;	       /* mmapped miniseed data */
    int64_t      mseedsz;	/* size of miniseed data */
    int64_t      mseedp;	/* current offset into mseed */
    int64_t      mseedn;	/* # actual data bytes in mseed */
    int64_t      cnt ; 	/* record count */

    /* miniseed configuration parameters */
    Stcflags        flags;	       /* configuration flags */
    int             frames_per_record,	/* frames per record */
		    firstframe,    /* first frame (of fpr) to fill */
		    set_data_offset,	/* user suggested offset of
					     * first data frame */
		    record_size;   /* size of a record in bytes */


    /* variables used internally by compression */
    int             nerr;	       /* number of errors encountered during
				    * (Steim 2) compression */
    int            *ffp;	       /* pointer to first frame of record */
    int            *x0,
		   *xn;	       /* pointer to first, last value of
				    * record */
    int            *fp;	       /* pointer to current frame (and
				    * therefore w0) */
    int             new_record_flag;	/* flag which is set to save
					     * the first sample in a
					     * record */
    int             fcnt;	       /* count of frames in current record */
    int            *wp;	       /* pointer to current w to fill */
    int             wcnt;	       /* count of w's filled in current
				    * frame */
    int            *data;	       /* buffer for returned data from
				    * ustc */
    long             datasz;	       /* size of data buffer */

    int             last_x;	       /* previous sample value */
    Tbl            *cbuf;	       /* buffer for samples submitted to
				    * csteim, but not yet packed into a
				    * word or frame. */
    int             maxclass;      /* maximum "class" for the samples in
				    * cbuf */
    int             code;	       /* accumulating code for current frame */
    Stcdef         *steimdef;      /* static table which defines
				    * parameters for compression
				    * algorithm */
    int             nsteimdef;     /* number of entries in steimdef */

    /* working variables of more general interest */
    long             nrecords,      /* count of records (= calls to
				    * save_record) */
		    n0,
		    n1;	       /* index of first and last sample in
				    * current record */
    unsigned char  *record;	       /* record buffer */
    int             recordsz;       /* size of record allotted */
    double          t0;	       /* first sample time for current data
				    * segment */

}               Msd;

#if 0
/* Steim compatibility */
#define Steim Msd 
#define SDH SDH1 
#define S1000 B1000 
#define S100 B100  

#define csteim(A,B,C,D)  cstc(A,B,C,D)
#define newsteim()       msdnew() 
#define usteim(A,B,C)    ustc(A,B,C)
#define freesteim(A)	 msdfree(A)
#define parse_seed_data_header(A)  msdhdr_unpack (A)
#endif

#ifdef	__cplusplus
extern          "C" {
#endif

extern Msd * msdnew ( void );
extern Msd_opaque * msdopaquenew ( void );
extern int cmsd ( Msd * msd, int (*save_record) (Msd *, long, long), int *data, long npts );
extern int cstc (Msd * msd, int (*save_record) (Msd *, long, long), int *data, long npts);
extern int dmsd ( FILE *file, Msd *msd, int verbose );
extern int dstc ( FILE *file, char *record, int record_size );
extern int msdclose(Msd *msd) ; 
extern int msdfix (Msd *msd) ;
extern int msdget ( Msd * msd, ... );
extern int msdhdr_unpack ( Msd * msd );
extern int msdnxt ( Msd *msd );
extern int msdpatch ( Msd *msd, int code, void *vnew );
extern int msdput ( Msd * msd, ... );
extern int msdrec ( Msd *msd, long rec );
extern int msdwrt ( Msd * msd, long n0, long n1 );
extern int samprate2seed ( double samprate, int *factor, int *multiplier );
extern int seed2samprate ( int factor, int multiplier, double *samprate );
extern int umsd ( Msd *msd, int **data, long *npts );
extern int unpack_sdh ( Msd * msd );
extern int ustc (Msd * msd, int **data, long *npts);
extern int ustc_patch ( Msd * msd, int **data, long *npts, char *path, int no_touch, int verbose );
extern void msd_free_opaque ( Msd_opaque *q ) ;
extern void msdfree ( Msd * msd );
extern void msdreinit ( Msd *msd );
extern void msdreinit(Msd *msd) ;
extern void pack_b100 ( Msd * msd, Stcflags flags, unsigned char *buffer, int *offset );
extern void pack_b1000 ( Msd * msd, Stcflags flags, unsigned char *buffer, int *offset );
extern void pack_msdhdr ( Msd * msd, long n0, long n1 );
extern void pack_sdh ( Msd * msd, Stcflags flags, unsigned char *buffer, int *offset );
extern void unpack_b100 ( Msd * msd, long offset );
extern void unpack_b1000 ( Msd * msd, long offset );
extern int unpack_b2000 (Msd * msd, long offset) ;

extern int msd_next_data_record ( Msd *msd );
extern int msdrec_offset ( Msd *msd, long offset );
extern int msd_find_data_record ( Msd *msd );

#ifdef	__cplusplus
}
#endif

#endif

